/*

  compile: gcc -O2 io_latency.c   # O2 needed?
  or:      gcc -O2 -fopenmp io_latency.c         # pathcc on SC
           OMP_NUM_THREADS=2 ./a.out ...
  
 Problem: Benchmarking external disks (RAID on controllers + SSD)
  isnt easy. IO-rate + stream depends on speed and num of disks.
  But latency is what a single VM in a bunch of VMs feels and what
  the quality defines. 
  Dont use buffers to measure IO because output to /dev/null 
  reduces speed (230MB/s on 4Gb-FC), direct IO is needed (400MB/s on 4Gb-FC).
  
  IO-Rate: max. num of random request per time summed over all threads,
    multiple threads read random, this fills command queue equaly
    and givesmaximum IO-rate (2012: 80..400 IO/s per DISK reading)
  Stream: max. num sequential Bytes per time (2012: 40..120MB/s per disk)
    do this using dd if=/dev/sda of=/dev/null flags=direct bs=1024k
  Latency: time range (min max) a single thread has to wait from 
    sending command and get the result (very system specifique!),
    this is not simply 1/IO-rate
    a slow RAID-controller is able to compensate low latency by
    more cache, getting same IO-rate like a fast controler 
  
  controller-example-read:
  - get read 1k command from host
  - reservation of 1k cache and sending read data command to disk (queue)
  - waits for the disk (but can accept more read requests to get more IO)
  - when data received, it sends data back to the host
  - waits for send finished (can do other IO)
  - frees cache for other requests
     
  ToDo: output %% of theoretical max defined by other parameters 
 
  on parallel:
    - IO: every thread reads random blocks from one central file
    - stream:  t = (blocksize/maxBW + latency)*numblks    
               # 42GB/s = 1MB/32us single thread
    - latency: (t1-t2) - (b1*n1-b2*n2)/maxBW = (n1-n2)*latency
                         b1*n1==b2*n2 same data volume choosen
               latency = (t1-t2) / (n1-n2) 
        
      dd /dev/zero                      thisPG /dev/zero
       10000*1MB  = 0.322..0.323s       r=0.314..0.323 w=0.003
       20000*.5MB = 0.327..0.329s       r=0.319..0.329 w=0.006
       20480000*512 = 6.98..7.05s       r=4.01..4.08 w=2.75..2.78
       latency = 0.005s / 10000 = 0.5us  maxIO=2e6/s
               = 6.66s / 20470000 = 0.325us
      # 1200 MHz = 1 / 0.833ns
      
  PC2010: 1GB     92 MB/s      132 IOPS 7.6ms (4kBsec,64MBcache,5400rpm,SATA2,1TB,NCQ)
       2t 1GB 75..92 MB/s 130..134 IOPS ???

  SC5832 1/500MHz = 2ns 4Gb-FC+VTrak-E310f+(5+1)RAID5(64k=128sec
    # 1MB==1e6B rs=read_sequential rr=read_random ws=write_sequential
    # m0n1:/dev/sda(5+1RAID5) rs=250MB/s rr=170/s
    io_latency(blk,num): m0n1.sda               ms0n1./dev/zero
          1048576  1000 ws5.79s rs10.30..10.33s   ws0.001s rs0.146s
           524288  2000 ws5.81s rs9.9..10.28s     ws0.003s rs0.273s             
           4096  256000 ws12.18s rs8.42s          ws0.339  rs0.771
         ca. ws170MB/s rs120MB/s  rs4.3us     
       /dev/zero: (1000*2**20B=0.146s, 2000*2**19B=0.273s) rs7GB/s rs1.6us  
    DIRECT 2097152  500  ws4.02s rs5.27s rs=250MB/s
           1048576  1000 ws4.11s rs4.12s rs=251MB/s 
            524288  2000 ws4.52s rs4.43s rs=240MB/s (340MB/s/2)
            16384  64000                         rr= (344/s/2)
            # ToDo: repeat using full disk capacity
            8192  128000 ws34.3 rs31.7s rr33MB/s rr=174=6*28/s (ahead=256)
            4096  256000 ws66.3 rs62.0s rr17MB/s rr=171=6*28/s
            2048  512000 ws132  rs117s  rr9MB/s  rr=169=6*28/s
            1024 1024000                      rr=172=6*28/s
            512  2048000                      rr=170=6*28/s 5.9ms (340/s/2)
        1x  1048576  1024 rs= 248 MB/s        rr= 169/s
        6x  1048576  1024 rs= 410 MB/s        rr= 624/s (iozone=6*164/s ??)
                          rs= 175..393 MB/s   rr= 601..627/s !!!
       12x  1048576  1024 rs= 141..384 MB/s   rr= 778..813/s file zu klein!
       24x  1048576  1024 rs= 127..215 MB/s   rr= 888..947/s file zu klein!
       12x  1048576 65536 rs= 233MB/s         rr= 613..656/s
  dd if=/dev/sda of=/dev/null bs=1024k count=1000 # 1GB/9.86s = 106MB/s
  dd if=/dev/sda of=/dev/null bs=1024k count=1000 iflag=direct # 4.18s 251MB/s
  dd of=/dev/sda if=/dev/zero bs=1024k count=1000 # 1GB/5.79s = 181MB/s
  dd of=/dev/sda if=/dev/zero bs=1024k count=1000 oflag=direct # 4.17s 252MB/s
  iozone -i 0 -i 1 -s 1048576 -t 1 -F /dev/sda # ws= 84MB/s rs=134MB/s
  ... -r 1024                                  # ws=169MB/s rs=111MB/s 1MB
  ... -r 1024 -I -i 2                          # ws=249MB/s rs=248MB/s rr=65/s
  ... -i 1 -s 1048576 -t 6 -r 1024 -I          # rs=400MB/s (6*1024*1MB 4Gb-FC)
  # threads * selbe file! Disk-cache + FC= 400MB/s
  ... -i 1 -s 1048576 -t 3 -r 1024 -I          # rs=398MB/s (3 threads, bs=1MB)
  ... -i 1 -s 1048576 -t 2 -r 1024 -I          # rs=373MB/s (2 threads, bs=1MB)
  ... -r 1    -I -i 2 -O                       # rr=1*168/s (262e3ops/26min)
  ... -r 1    -I -i 2 -O -t 6                  # rr=6*164/s (4MB/s/4k=1kIOPs)
  # skaliert ideal? wegen gleichen + kleinen files + caches?
  # zu kleine files liefern astronomische IOPs (cache!):
  ... -i 2 -s 131072  -t 12 -r 4 -I -O -p 100  # rr=12*884/s (44MB/s/4k=10kIOPs)
  ... -i 2 -s 131072  -t 6  -r 4 -I -O -p 100  # rr=6*1756/s (40MB/s)
  ... -i 2 -s 131072  -t 1  -r 4 -I -O -p 100  # rr=3882/s (33e3ops/8.4s)
  ... -i 2 -s 262144  -t 1  -r 4 -I -O -p 100  # rr=407/s (66e3ops/161s,1.2MB/s/4k=400IOPs)
  ... -i 2 -s 524288  -t 1  -r 4 -I -O -p 100  # rr=195/s (131e3ops/11.2m)
  ... -H -O -p 100
  # selbe file random-muster threads verschieden?
  m4n1:/dev/sdi(12raid0)
   -t1  1048576  1024 rs=109MB/s  rr=220/s (zu klein, 12*128MB?=1.4GB)
   -t1  1048576  4096 rs=106MB/s  rr=170/s
   -t1  1048576  8192 rs=106MB/s  rr=160/s
   -t1  1048576 16384 rs=100MB/s  rr=157/s 0.2% of disk space
   -t1  1048576 32768 rs=104MB/s  rr=152/s (read ahead schaedlich? 0.3%)
   -t12 1048576 32768 rs=404MB/s  rr=1046/s (nur 0.3% der Disk)
   -t6  1048576     1 rr=8630-8747/s
   -t12 1048576     1 rr=9473-9564/s
   -t18 1048576     1 rr=9142-9213/s
   -t36 1048576     1 rr=9073-9495/s (system limit?)
   -t1           12TB rr=120/s  8.3ms (100% of (12+0)raid0) !!!
   -t6           12TB rr=563-572/s   (6*1.78ms=10.7ms)
   -t12          12TB rr=844/s       (12*1.2ms=14ms)
   -t24          12TB rr=1080/s      (24*0.93ms=22ms)
   -t30          12TB rr=1128-1136/s (30*0.89ms=27ms)
   -t36          12TB rr=1154-1185/s (36*0.85ms=30ms/task,12*0.85ms=10.2ms/dsk)
   -t42          12TB rr=1158-1199/s
  
  pse:/dev/sdc=SSD_INTEL_SSDSC2MH120A2 120GB
   -t1   1048576 16384  rs=240MB/s rr=5530/s(4k)  (rs.lat=2.3us(512k-256k))
   -t1   1048576  8192  rs=244MB/s rr=5530/s(4k)  lat=182us
   -t2   1048576  8192  rs=286MB/s rr=8570/s(4k)  lat<=117us*2=233us
   -t4   1048576  8192  rs=285MB/s rr=1.38e4/s(4k) lat<=72us*4=290us
   -t6   1048576  8192  rs=285MB/s rr=1.69e4/s(4k) 
   -t8   1048576  1024  rs=283MB/s rr=1.88e4/s(4k)
   -t10  1048576  1024  rs=285MB/s rr=2.04e4/s(4k)
   -t12  1048576  8192  rs=286MB/s rr=2.05e4/s(4k) lat<=50us*12 max_rr
   -t14  1048576  1024  rs=285MB/s rr=2.13e4/s(4k)
   -t16  1048576  1024  rs=285MB/s rr=2.12e4/s(4k)
   -t18  1048576  1024  rs=285MB/s rr=1.85e4/s(4k)
   -t24  1048576  8192  rs=286MB/s rr=1.90e4/s(4k)
    
  vtrak8 Peter 16.07.2012 100%=1.1TB (4+1)RAID5
   -t8 rr=1100/s(5*0.9ms=4.5ms/dsk)  13-15ms lt. status-website
   -t1 rr=184/s(5.4ms)                4-6ms  lt. status-website iozone=178/s
   
  meggie./dev/md3(4raid0) SATA 
   md3 : active raid0 sdh1[3] sdg1[2] sdf1[1] sde1[0] 512k  540MB/s
   -t1     4TB rr=100/s(10ms) (cpu=0.8%, 100%cpu load by other users)
   -t2     4TB rr=171/s
   -t4     4TB rr=252/s
   -t8     4TB rr=316/s  (cpu=10%)
   -t16    4TB rr=360/s
   -t32    4TB rr=408/s(4*2.5ms) (cpu=53%,  100%cpu load by other users)
   -t64    4TB rr=436/s (cpu=0.7%)
   -t1     1MB rr=7397-8141/s (4k, 8300/s 512B-1K, cpu=40%)
   -t2     1MB rr=10160-11320/s
   -t8     1MB rr=12340-13640/s
   -t16    1MB rr=10250-11600/s (4k, 8160-8500/s 512B-1K, cpu=500%)
   md3=(4+0)raid0 64k-chunks (-c 64) 583-598MB/s (-c 16 420-490MB/s)
   # (-c 1024 237MB/s, -c 512 426-464MB/s, -c 128 587MB/s)
   -t1     4TB rr=100/s wr=796/s 4k  (775/s -c 4 552MB/s, 785/s -c 16, 795/s -c 512)
   -t2     4TB rr=168/s wr=768/s 4k (wrmax=829/s 32k)
   -t4     4TB rr=250/s wr=756/s 4k (-c 128 243/s)
   -t8     4TB          wr=790/s 4k
   -t16    4TB          wr=768/s 4k
   md3=(3+1)raid5 64k-chunks, algorithm 2 [4/4] [UUUU] ca.2h build
   -t1     3TB rr=95/s  10.5ms  wr=66/s 4k-16k (40/s .5-2k) (1MB rr=157e3/s wr=39e3/s 64k-chunk)
   -t2     3TB rr=162/s         wr=71/s 4k-16k (62/s 1k user-nice19-cpu100%)
   -t4     3TB rr=236-237/s     wr=91-93/s 4k-8k (86-88/s 1k-2k,86-91/s 16k-64k)
   -t8     3TB rr=298/s 4k cpu=17%
   -t16    3TB rr=343/s
   -t32    3TB rr=383/s 4k,8k   wr=132/s 4k (120/s 1k)
   -t64    3TB rr=423/s 1k-4k   wr=129-134/s 4k-16k (124/s .5-2k) cpu=100%
   md2=(2raid0) wird benutzt?  readahead=4096=2M 512k-chunks
   md2 : active raid0 sdd1[1] sdc1[0] 512k  278..295MB/s (ahead hat kein effekt)
   -t1     2TB rr=89/s (cpu=0.6%)
   -t2     2TB rr=123/s (cpu=0.9%)
   -t4     2TB rr=142/s     (readahead=256,128) auch waehrend md3-build
   -t4     2TB rr=149/s     (readahead=8) 298MB/s
   -t32    2TB rr=166-171/s (cpu=22%) (readahead=256, ToDo: set0?)

 quantum.5TB (5+1)raid5 dd.direct 327-349MB/s (==sda 1.7TB)
   sda=330-349MB/s sdb=316-399MB/s sdc=291-304MB/s
   -t1 rr=147/s 512B  -t8 rr=600/s 4k   

 comp2.sdb.15TB dd.direct= 960-983MB/s (8+2)raid6(a2TB SATA via LSI-Controller)
  # dd.direct.2M 1.0GB/s 4M=1.1GB/s 8M=1.1GB/s 16M=763-796MB/s
    -t1 rr=135/s 2k  -t 8 rr=759/s 2k  -t 16 rr=1117/s 4k  -t 32 rr=1537/s 4k
  ToDo: check config!  
  
 */

#define _GNU_SOURCE    // adds O_DIRECT = 040000 = bit14

#include <omp.h>  // not available on MIPS-gcc-4.1.2 (sc) !? use pathcc
// /usr/lib/gcc/x86_64-redhat-linux/4.1.1/include/omp.h
// /usr/include/pthread.h
// /usr/share/PDT/include/kai/omp.h
// /usr/share/PDT/include/kai/pthread.h

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
// #define __USE_GNU   // adds O_DIRECT = 040000 = bit14
// #define __USE_LARGEFILE64 // adds O_LARGEFILE 
#include <fcntl.h>
//#include <asm/fcntl.h>

#include <string.h>
#include <time.h>
#include <sys/time.h> // gettimeofday()
#define E0(xx) {puts("ERR: "xx);} // puts adds an \n


int main( int argn, char *argv[]) {
  int minrbsize=512;    // min random blocksize (512)
  int maxrbsize=4*4096; // max random blocksize (4*4096)
  // blocksize= max sequential blocksize
  int i, fd=0, blocksize=1024*1024, blocknum=0, writemode=0;
  off_t *buf2=NULL;
  char crc=0, *buf=NULL, *fname="/dev/zero";
  double t3;
  struct timeval t1, t2; /* tv_sec tv_usec */
  int oflag = O_DIRECT | O_LARGEFILE;

  if (argn>1 && argv[1][0]=='w' && argv[1][1]=='\0') writemode=1;
  if (argn>3) blocksize=atoi(argv[argn-3]);
  if (argn>2) blocknum=atoi(argv[argn-2]);
  if (argn>1) fname=argv[argn-1]; /* last argument is file*/

  printf("io_latency v2012-07 Joerg Schbg.\n"
         " (max)blocksize=%d (min)blocknum=%d file=arg1=%s writemode=%d\n",
          blocksize, blocknum, fname, writemode);
  if (argn<3) printf(
    " args:    [w] [[maxblocksize] filesize/maxblocksize] file/devicename\n"
    " sample:  OMP_NUM_THREADS=8 ./a.out /dev/sdb # random read full disk\n");

  // get max disk/file space using lseek (works also for /dev/sda)
  printf(" getting filesize ...\n");
  fd=open(fname, O_RDONLY); if(fd<=0){E0("open.w");}
  if (fd) {
    off_t i1;
    i1=lseek(fd,0,SEEK_END); // needed for multithread
    if (i1==(off_t)-1) E0("lseek");
    if (i1!=(off_t)-1)
      printf(" got filesize %.9e (GB=1e9 TB=1e12) bit0-9,10-19= %d %d\n",
             1.*i1, (int)(i1&1023), (int)((i1>>10)&1023));
    if (!blocknum) blocknum= i1 / blocksize;  // autoset
    close(fd); fd=0;
  }
  if (!blocknum) blocknum=1024;
  printf(" blocknum= %d\n", blocknum);

#define MAX_THREADS 64  // 2^x, used to alloc enough buffers, 2...2^31
  // size=off_t*STORED_RNDS=8*16MB=128MB, maxfilelen=STORED_RNDS*.5kB=8GB
#define STORED_RNDS (1<<24) 
  printf(" start array initialization, RAND_MAX= %.1e ...\n",1.*RAND_MAX);
  // buf=malloc((size_t)blocksize+4096); if (!buf) {E0("malloc");exit(1);}
  // not alowed: buf2=((buf>>12)+1)<<12; // align to 4kB
  /* memalign is needed for O_DIRECT */
  posix_memalign((void **)&buf ,4096,blocksize*MAX_THREADS);
  posix_memalign((void **)&buf2,4096,sizeof(off_t)*STORED_RNDS);
  if ((!buf) || (!buf2)) E0("malloc");
  // printf(" buf= %p %p\n", buf, buf2);
  // free(buf);
  // initialize mem to avoid page initialization within measurement
  memset(buf,0,blocksize*MAX_THREADS); // ToDo: thread local init
  // generate (slow) random sequence for full storage size / 512 B
  // minrbsize=512; maxrbsize=4096;  // random blocksize
  srandom(getpid());
  // multiple of 512 (9bit) within file
  for (i=0;i<STORED_RNDS;i++)
    buf2[i]=(((off_t)random())<<9) % (blocknum*(off_t)blocksize);
    

#if 0
  printf(" start sequential write ...\n");
  for(;;){
    // do not use perror, use strace to debug (less overhead)
    fd=open(fname, O_WRONLY|O_CREAT|oflag);
    if (fd<=0) { E0("open.w, try -oflag");
      fd=open(fname, O_WRONLY|O_CREAT); if(fd<=0){E0("open.w");break;}
    }
    gettimeofday(&t1,NULL);
    for (i=0;i<blocknum;i++) {
      if (blocksize!=write(fd,buf,blocksize)) {E0("write");exit(1);}
      // blocksize > ... may fail for O_DIRECT ???
    }
    gettimeofday(&t2,NULL);
    t3= t2.tv_sec+1e-6*t2.tv_usec
       -t1.tv_sec-1e-6*t1.tv_usec;
    printf(" wseq blk= %7d size= %.2e t/s= %6.3f   %.3e B/s\n", 
           blocksize, blocknum*1.*blocksize, t3, blocknum*1.*blocksize/t3);
    close(fd);fd=0;
    // 2012-07 /dev/zero io 1000e3MB/0.15s, dd=100e3MB/3.065s  34GB/s
    break;
  }
#endif

#if 0
  printf(" start sequential read ...\n");
  {
    int bsize; off_t i1; // int pt_i;
    fd=open(fname, O_RDONLY|oflag); 
    if (fd<=0) { E0("open.r, try -oflag ...");
      fd=open(fname, O_RDONLY); if (fd<=0) E0("open.r");
    }
    for (bsize=blocksize;bsize>=blocksize/4;bsize/=2) {
      gettimeofday(&t1,NULL);
      // static means: trhread0=0..19, thread1=20..29 ... thread4=80..99
// ToDo: teile in 32 grosse blocks, mache einmal lseek, dann nur lesen
//  kleinere 2**x-wachsende chunks, eher abbrechen (schneller messen)?
//  braucht nicht so riesige files, aber random braucht grosse files
#pragma omp parallel for private(i1) schedule(static,32)
      for (i1=0;i1<blocknum*(off_t)blocksize/bsize;i1++) {
        // pt_i=omp_get_thread_num();
        lseek(fd,i1*(off_t)bsize,SEEK_SET); // needed for multithread
        // every thread (max 64) has its own buffer
        if (bsize!=read(fd,buf+blocksize*(i1&(MAX_THREADS-1)),bsize))
          {E0("read");exit(1);}
      }
      gettimeofday(&t2,NULL);
      t3= t2.tv_sec+1e-6*t2.tv_usec
         -t1.tv_sec-1e-6*t1.tv_usec;
      printf("  rseq blk= %7d size= %.2e t/s= %6.3f   %.3e B/s\n", 
            bsize, blocknum*1.*blocksize, t3, blocknum*1.*blocksize/t3);
    }
    close(fd);fd=0;
    // 2012-07 /dev/zero io 100e3MB/3.064s, dd=100e3MB/3.065s  34GB/s
  }
#endif

  if (writemode)
  {
    int bsize, j, irnd;
    printf(" start random write ...\n");
    fd=open(fname, O_WRONLY|O_CREAT|oflag); 
    if (fd<=0) { E0("open.w, try -oflag ...");
      fd=open(fname, O_WRONLY|O_CREAT); if (fd<=0) E0("open.w");
    }
    for (irnd=0,bsize=minrbsize;bsize<=maxrbsize;bsize*=2) {
      for (j=64;j<=4096;irnd+=j,j*=2) {
        gettimeofday(&t1,NULL);
        // random in multithread?
#pragma omp parallel for private(i,t3) schedule(static,1)
        for (i=0;i<j;i++) {
          // bsize must be 2^x for correct down rounding
          //  bad parallel performance for interleaved access?
          lseek(fd,(buf2[(i+irnd)&(STORED_RNDS-1)]&~(off_t)(bsize-1)),SEEK_SET);
          if (bsize!=write(fd,buf+bsize*(i&(MAX_THREADS-1)),bsize)) {E0("write");exit(2);}
        } // i
        gettimeofday(&t2,NULL);
        t3= t2.tv_sec+1e-6*t2.tv_usec
           -t1.tv_sec-1e-6*t1.tv_usec;
        if (t3>8. || j>=4096) { // 8 16 32 64 (4 outputs)
          printf("  wrnd blk= %7d size= %.2e t/s= %6.3f   ops= %4d  %.3e IOPS\n",
            bsize, blocknum*1.*blocksize, t3, j, j*1./t3);
          if (t3>60) break; // max. 2min
        } // if 8..60s
      } // j
    } // bsize
  } // rrnd

  if (!writemode)
  {
    int bsize, j, irnd;
    printf(" start random read ...\n");
    fd=open(fname, O_RDONLY|oflag); 
    if (fd<=0) { E0("open.r, try -oflag ...");
      fd=open(fname, O_RDONLY); if (fd<=0) E0("open.r");
    }
    for (irnd=0,bsize=minrbsize;bsize<=maxrbsize;bsize*=2) {
      for (j=64;j<=4096;irnd+=j,j*=2) {
        gettimeofday(&t1,NULL);
        // random in multithread?
#pragma omp parallel for private(i,t3) schedule(static,1)
        for (i=0;i<j;i++) {
          // bsize must be 2^x for correct down rounding
          //  bad parallel performance for interleaved access?
          lseek(fd,(buf2[(i+irnd)&(STORED_RNDS-1)]&~(off_t)(bsize-1)),SEEK_SET);
          if (bsize!=read(fd,buf+bsize*(i&(MAX_THREADS-1)),bsize)) {E0("read");exit(2);}
        } // i
        gettimeofday(&t2,NULL);
        t3= t2.tv_sec+1e-6*t2.tv_usec
           -t1.tv_sec-1e-6*t1.tv_usec;
        if (t3>8. || j>=4096) { // 8 16 32 64 (4 outputs)
          printf("  rrnd blk= %7d size= %.2e t/s= %6.3f   ops= %4d  %.3e IOPS\n",
            bsize, blocknum*1.*blocksize, t3, j, j*1./t3);
          if (t3>60) break; // max. 2min
        } // if 8..60s
      } // j
    } // bsize
  } // rrnd
  if (fd) close(fd); fd=0;
  if (buf) free(buf); if (buf2) free(buf2);
  return ((crc!=17)?0:1);
}
